#
# undo.tcl - implement undo function
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Dec, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: undo.tcl,v 1.2 2005/01/09 21:00:21 tronkko Exp $
#
# $Log: undo.tcl,v $
# Revision 1.2  2005/01/09 21:00:21  tronkko
# updated comments
#
# Revision 1.1  2004/12/09 22:15:05  tronkko
# Initial revision
#
#
#


#
# Undo changes up to previous undo limit.  Updates views as necessary.
#
proc undo {} {
  global undo_x
  global undo_y
  global undo_start
  global undo_ptr
  global undo_chip
  global undo_workspace
  global undo_len
  global workspace_chips
  global editor_widget
  global selector_widget

  # find backward to previous undo limit
  set limit $undo_ptr
  while { $limit != $undo_start } {
    
    # decrease ptr
    if { $limit != 0 } {
      incr limit -1
    } else {
      set limit [expr $undo_len - 1]
    }

    # is limit?
    if { $undo_workspace($limit) < 0 } {
      # found a limit
      break
    }
  }

  # undo changes since previous undo limit
  if { $limit != $undo_start } {

    # undo events backwards
    set i $undo_ptr
    set count 0
    while { $i != $limit } {    
      # decrease ptr
      if { $i != 0 } {
        incr i -1
      } else {
        set i [expr $undo_len - 1]
      }

      # restore chip
      set cur $undo_workspace($i)
      if { $cur >= 0 } {
        set x   $undo_x($i)
        set y   $undo_y($i)
        set workspace_chips($cur,$x,$y) $undo_chip($i)
        incr count
      }

    }

    # fix undo buffer for further undo
    set undo_ptr $limit

    # restore edit status
    set_edited $undo_chip($limit)

    # Refresh views.  Refreshing two two views completely takes some time
    # but undo is used quite rarely.
    refresh_view $editor_widget
    refresh_view $selector_widget
    refresh_title
    status "Restored $count chips from undo buffer"
  } else {
    # undo buffer empty/no complete action in buffer
    bell
    status "No more undo"
  }
}


#
# Mark undo limit.  That is, store situation into which the very next
# undo command will return.
#
proc undo_limit {} {
  global undo_x
  global undo_y
  global undo_start
  global undo_ptr
  global undo_chip
  global undo_workspace
  global edit_status

  # mark limit
  set undo_x($undo_ptr) -1
  set undo_y($undo_ptr) -1
  set undo_workspace($undo_ptr) -1
  set undo_chip($undo_ptr) $edit_status

  # advance undo ptr
  push_undo
}


#
# Edit chip at (x,y) in ith workspace, and save undo information
#
proc set_chip { i x y value } {
  global undo_x
  global undo_y
  global undo_start
  global undo_ptr
  global undo_chip
  global undo_workspace
  global workspace_chips

  # save original chip
  set undo_x($undo_ptr) $x
  set undo_y($undo_ptr) $y
  set undo_workspace($undo_ptr) $i
  set undo_chip($undo_ptr) $workspace_chips($i,$x,$y)

  # modify chip
  set workspace_chips($i,$x,$y) $value

  # advance undo ptr
  push_undo
}


#
# Advance undo pointer.
#
proc push_undo {} {
  global undo_ptr
  global undo_len
  global undo_start

  # advance pointer
  incr undo_ptr
  if { $undo_ptr >= $undo_len } {
    set undo_ptr 0
  }

  # erase old undo info
  if { $undo_ptr == $undo_start } {    
    incr undo_start
    if { $undo_start >= $undo_len } {
      set undo_start 0
    }
  }
}


#
# Reset undo buffer.  The function should be used after deleting or
# changing workspaces.
#
proc reset_undo {} {
  global undo_start
  global undo_ptr
  
  set undo_start 0
  set undo_ptr 0
  undo_limit
}