#
# view.tcl - view widget with workspaces
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Dec, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: view.tcl,v 1.6 2005/01/16 23:39:48 tronkko Exp $
#
# $Log: view.tcl,v $
# Revision 1.6  2005/01/16 23:39:48  tronkko
# zoom_levels is a list, zoom is index to that list
#
# Revision 1.5  2005/01/09 20:56:18  tronkko
# added edit mode, only update changed chips to speed up undo, speed up
# drawing by not changing cursor unless absolutely necessary
#
# Revision 1.4  2004/12/09 22:13:51  tronkko
# initial support for showing regions, numerous small fixes
#
# Revision 1.3  2004/12/01 20:12:58  tronkko
# clarified the role of palette and workspace by renaming some
# variables, show drop shadow in draw mode
#
# Revision 1.2  2004/11/30 20:27:48  tronkko
# draw with selected chips, mouse button 1 operates according to
# mouse_mode, copy between workspaces
#
# Revision 1.1  2004/11/28 11:08:21  tronkko
# Initial revision
#
#
#

# Coordinate of initial chip.  The variable is used for selecting multiple
# chips using mouse.
set view_x0 -1
set view_y0 -1

# Offset to upper left corner when moving multiple chips.  The variable
# lets the user drags a selection by a chip.
set view_fx -1
set view_fy -1

# initial coordinates for panning
set view_pan_x0 -1
set view_pan_y0 -1
set view_pan_xview 0
set view_pan_yview 0

# drag command, works like mouse_mode
set view_drag_op "none"

# status of drag operation: -1 denied, 0=allowed, 1=under way
set view_dragged -1

# currently highlighted region (for updating drop highlight)
set view_selection ""



#
# Create view widget for organizing chips.  Returns the path name that
# can be used for operating the widget.
#
proc create_view { win } {
  global zoom
  global cur_workspace
  global write_protect
  global view_border
  
  # name of canvas object
  set c $win.canvas

  # set up defaults
  set cur_workspace($c) -1
  set write_protect($c) 0
  set view_border($c) 1
  
  # frame surrounding the canvas and scroll bars
  frame $win -bd 0
 
  # canvas containing chips
  canvas $c -width 320 -height 256 \
      -xscrollcommand "$win.xbar set" \
      -yscrollcommand "$win.ybar set" \
      -takefocus 1 -bd 2 -bg black -relief sunken -highlightthickness 0
  scrollbar $win.xbar -orient horizontal -command "$c xview" -takefocus 0
  scrollbar $win.ybar -orient vertical   -command "$c yview" -takefocus 0
  
  # pack canvas and scrollbars
  grid $c -row 0 -column 0 -sticky nsew -ipadx 2 -ipady 2
  grid $win.ybar -row 0 -column 1 -sticky ns
  grid $win.xbar -row 1 -column 0 -sticky ew
  grid rowconfigure $win 0 -weight 1
  grid columnconfigure $win 0 -weight 1    
  pack $win -fill both -anchor nw -expand yes -padx 6 -pady 6

  # for highlighing chip under cursor
  $c bind "chip" <Enter> "view_enter $c %X %Y"
  $c bind "select"  <Enter> "view_enter $c %X %Y"
  $c bind "highlight" <Leave> "view_leave $c %X %Y"
  bind $c <Leave> "view_leave $c %X %Y"

  # move highlight as the mouse pointer is moved (this is important when 
  # more than one chips is selected)
  $c bind "highlight" <Motion> "view_enter $c %X %Y"
  $c bind "region" <Motion> "view_enter $c %X %Y"

  # select, drag and drop
  bind $c <1> "view_action $c %X %Y"
  bind $c <Shift-1> "view_action $c %X %Y select"
  bind $c <Control-1> "view_action $c %X %Y move"
  bind $c <B1-Motion> "view_drag $c %X %Y"
  bind $c <ButtonRelease-1> "view_drop $c %X %Y"
  bind $c <2> "view_action $c %X %Y draw"
  bind $c <B2-Motion> "view_drag $c %X %Y"
  bind $c <ButtonRelease-2> "view_drop $c %X %Y"
  
  # panning
  bind $c <3> "view_action $c %X %Y pan"
  bind $c <B3-Motion> "view_drag $c %X %Y"
  bind $c <ButtonRelease-3> "view_drop $c %X %Y"
  bind $c <Control-3> "view_action $c %X %Y edit"

  # cursor movement
  bind $c <KeyPress-Left>  "view_left  $c"
  bind $c <KeyPress-Right> "view_right $c"
  bind $c <KeyPress-Up>    "view_up    $c"
  bind $c <KeyPress-Down>  "view_down  $c"
  bind $c <Shift-KeyPress-Left>  "view_expand_left  $c"
  bind $c <Shift-KeyPress-Right> "view_expand_right $c"
  bind $c <Shift-KeyPress-Up>    "view_expand_up    $c"
  bind $c <Shift-KeyPress-Down>  "view_expand_down  $c"

  # set cursor according to mode
  restore_cursor $c

  # return name of the widget
  return $c
}


#
# Set border width.  Set w to zero to disable border.
# 
proc set_border { win w } {
  global view_border
  if { $view_border($win) != $w } {
    set view_border($win) $w
    $win delete all
    refresh_view $win
  }
}


#
# Select workspace to show on a view widget
#
proc select_workspace { win i } {
  global cur_workspace
  global workspace_xview
  global workspace_yview

  # get id of currently visible workspace
  set cur $cur_workspace($win)

  # save view coordinates of the currently visible workspace
  if { $i != $cur } {
    set workspace_xview($cur) [lindex [$win xview] 0]
    set workspace_yview($cur) [lindex [$win yview] 0]
    $win xview moveto 0
    $win yview moveto 0
  }

  # switch to new workspace
  set cur_workspace($win) $i
  
  # paint chips into view
  $win delete all
  refresh_view $win
  
  # fetch previous view coordinates if changing workspaces
  if { $i != $cur } {
    $win xview moveto $workspace_xview($i)
    $win yview moveto $workspace_yview($i)
  }
}


#
# Update view window after changing zoom level or switching to another
# workspace.  Note that the function does not completely erases the view.
# Be sure to delete all chips by "$win delete all" if changing view
# or re-sizing it.
#
proc refresh_view { win } {
  global nworkspaces
  global workspace_chips
  global workspace_width
  global workspace_height
  global chip_image
  global zoom
  global chipw
  global chiph
  global cur_workspace
  global src_workspace
  global view_border
  global workspace_nregions
  global workspace_region_used
  global workspace_region_x
  global workspace_region_y
  global workspace_region_width
  global workspace_region_height
  global workspace_region_name
  global region_font
  global zoom_levels

  # get border width
  set bw $view_border($win)

  # get current zoom level
  set z [lindex $zoom_levels $zoom]

  # compute pixel dimensions of a chip (including border)
  set w [expr $chipw*$z + $bw]
  set h [expr $chiph*$z + $bw]
  
  # insert chips into widget
  set cur $cur_workspace($win)
  if { $cur >= 0 } {
    # insert chips
    for {set y 0} {$y < $workspace_height($cur)} {incr y} {
      for {set x 0} {$x < $workspace_width($cur)} {incr x} {
        # get currently visible image
        set old [$win itemcget "chip_${x}_${y}" -image]

        # get new image
        set chip $workspace_chips($cur,$x,$y)
        set img $chip_image($chip,1)

        # Insert image to canvas if the image changes.  Doing this speeds
        # up undo considerably.
        if { [string compare $old $img] != 0 } {
          $win delete "chip_${x}_${y}"
          set pos [view_pos $win $x $y 1 1]
          set x1 [expr [lindex $pos 0] + $bw]
          set y1 [expr [lindex $pos 1] + $bw]
          $win create image $x1 $y1 -anchor nw \
              -tag "chip_${x}_${y}" -image $img
          $win addtag "chip" withtag "chip_${x}_${y}"
        }
      }
    }

    # insert regions    
    for {set i 0} {$i < $workspace_nregions($cur)} {incr i} {
      if { $workspace_region_used($cur,$i) } {
        # compute position of region in screen coordinates
        set pos [view_pos $win \
            $workspace_region_x($cur,$i) \
            $workspace_region_y($cur,$i) \
            $workspace_region_width($cur,$i) \
            $workspace_region_height($cur,$i)]
        set x1 [lindex $pos 0]
        set y1 [lindex $pos 1]
        set x2 [lindex $pos 2]
        set y2 [lindex $pos 3]
        
        # paint frame around the region
        $win create rectangle $x1 $y1 $x2 $y2 -tag "frame$i" \
            -width 2 -outline white
        $win create polygon \
            $x2 $y2 $x2 [expr $y2 - 14] [expr $x2 - 14] $y2 \
            -tag "resize$i" \
            -width 2 -fill white
        
        # position name above the frame
        set xt [expr $x1 + 2]
        set yt [expr $y1 - 2]
        
        # paint name
        set text $workspace_region_name($cur,$i)
        set len [string length $text]
        $win create rectangle $x1 $y1 [expr $x1+$len*8] [expr $y1-16] \
             -tag "region$i" -width 2 -fill white -outline white
        $win create text $xt $yt -anchor sw -tag "text$i" \
            -text $text -fill black -font $region_font
        
        # edit region by double clicking the name
        $win bind "region$i" <Double-1> "edit_region $cur $i"
        $win bind "text$i" <Double-1> "edit_region $cur $i"

        # highlight region name when mouse cursor enters the name
        $win bind "region$i" <Enter> "enter_region $win $i %X %Y"
        $win bind "region$i" <Leave> "leave_region $win $i"
        $win bind "text$i" <Enter> "enter_region $win $i %X %Y"
        $win bind "text$i" <Leave> "leave_region $win $i"

        # move frame
        $win bind "region$i" <1> "puts \"1: $cur $i %X %Y\""
        $win bind "text$i" <1> "puts \"1: $cur $i %X %Y\""
        $win bind "region$i" <ButtonRelease-1> "puts \"rel1: $cur $i %X %Y\""
        $win bind "text$i" <ButtonRelease-1> "puts \"rel1: $cur $i %X %Y\""
      }
    }
    
    # set color of area showing through gridlines
    $win configure -bg black
  } else {
    # invalid workspace
    $win configure -bg grey50
  }
  
  # re-define scrolling region and unit size
  $win configure -scrollregion [$win bbox "chip"]

  # highlight selection
  if { $cur == $src_workspace  &&  $cur >= 0 } {
    global view_x1
    global view_y1
    global view_x2
    global view_y2
    view_select $win $view_x1 $view_y1 $view_x2 $view_y2
  }  
}


#
# Create visual clues to indicate that something happens if pressing
# left mouse button abouve region title.
# 
proc enter_region { win i x y } {
  view_leave $win $x $y
  $win itemconfigure "text$i" -fill red
  $win configure -cursor hand2
}


#
# Mouse pointer leaves region title.
#
proc leave_region { win i } {
  $win itemconfigure "text$i" -fill black
  restore_cursor $win
}


#
# Erase selection from either selector or editor window.
#
proc erase_chips {} {
  global selection_owner
  global view_x1
  global view_y1
  global view_x2
  global view_y2
  global src_workspace
  global chip_image
  global zoom
  global write_protect

  if { $selection_owner != "" } {
    # has selection
    if { ! $write_protect($selection_owner) } {
      undo_limit
      set count 0
      for {set y $view_y1} {$y < $view_y2} {incr y} {
        for {set x $view_x1} {$x < $view_x2} {incr x} {
          # replace the chip with an invalid chip
          set_chip $src_workspace $x $y 0

          # update chip on screen
          $selection_owner itemconfigure "chip_${x}_${y}" \
              -image $chip_image(0,1)

          incr count
        }
      }

      # file has now been edited
      set_edited 1
      status "Erased $count chips"
    } else {
      modal_dialog "Error" "The view has been locked against accidental modifications.  Please switch off the protection in order to delete chips." error
    }
    
  } else {
    status "Nothing selected"
  }
}


#
# Select chip at left.  The function is invoked by pressing left cursor
# key.
#
proc view_left { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1
  global cur_workspace
  global src_workspace
  global selection_owner

  # Get name of view widget owning the selection.  Not using the
  # the window for which the event is triggered makes it easy to move
  # cursor with cursor keys while painting over some other widget using
  # mouse.  This makes painting easier for user, as he does not need to
  # move the mouse pointer to the other window just to select a chip -
  # he can use the cursor keys for that.
  set win $selection_owner
  if { $win == "" } return 

  # get visible workspace of the widget with selection
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    # cursor is in visible workspace
    if { $view_x1 > 0 } {
      # modify selection
      set x1 [expr $view_x1 - 1]
      set y1 $view_y1
      set x2 [expr $x1 + 1]
      set y2 [expr $y1 + 1]
      view_select $win $x1 $y1 $x2 $y2

      # bring selection into screen
      view_show_selection $win

      # highlight drop area in window containing the mouse cursor
      view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
    }
  } else {
    if { $src_workspace != $cur } {
      # selection is in buried workspace, activate current workspace
      reset_selection $win
    }
  }
}


#
# Move cursor right.  This and the following function are more or less
# copies of the view_left function, so the comments are omitted.
#
proc view_right { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1
  global workspace_width
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 

  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set width [expr $workspace_width($cur) - 1]
    if { $view_x1 < $width  &&  $view_x1 >= 0 } {
      set x1 [expr $view_x1 + 1]
      set y1 $view_y1
      set x2 [expr $x1 + 1]
      set y2 [expr $y1 + 1]
      view_select $win $x1 $y1 $x2 $y2
      view_show_selection $win
      view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}

proc view_up { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    if { $view_y1 > 0 } {
      set x1 $view_x1
      set y1 [expr $view_y1 - 1]
      set x2 [expr $x1 + 1]
      set y2 [expr $y1 + 1]
      view_select $win $x1 $y1 $x2 $y2
      view_show_selection $win
      view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}

proc view_down { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1
  global workspace_height
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set height [expr $workspace_height($cur) - 1]
    if { $view_y1 < $height  &&  $view_y1 >= 0 } {
      set x1 $view_x1
      set y1 [expr $view_y1 + 1]
      set x2 [expr $x1 + 1]
      set y2 [expr $y1 + 1]
      view_select $win $x1 $y1 $x2 $y2
      view_show_selection $win
      view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}


#
# Expand/shrink current selection.
#
proc view_expand_left { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set x1 $view_x1
    set y1 $view_y1
    if { $x1 >= 0 } {
      set x2 [expr $view_x2 - 1]
      set y2 $view_y2
      if { $x1 < $x2 } {
        view_select $win $x1 $y1 $x2 $y2
        view_show_selection $win
        view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
        status ""
      } else {
        status "Cannot shrink the selection any more"
      }
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}

proc view_expand_right { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1 
  global view_x2
  global view_y2 
  global workspace_width
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set x1 $view_x1
    set y1 $view_y1
    set width $workspace_width($cur)
    if { $x1 < $width  &&  $x1 >= 0 } {
      set x2 [expr $view_x2 + 1]
      set y2 $view_y2
      if { $x2 <= $width } {
        view_select $win $x1 $y1 $x2 $y2
        view_show_selection $win
        view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
        status ""
      } else {
        status "Right edge reached"
      }
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}

proc view_expand_up { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set x1 $view_x1
    set y1 $view_y1
    if { $y1 >= 0 } {
      set x2 $view_x2
      set y2 [expr $view_y2 - 1]
      if { $y1 < $y2 } {
        view_select $win $x1 $y1 $x2 $y2
        view_show_selection $win
        view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
        status ""
      } else {
        status "Cannot shrink the selection any more"
      }
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}

proc view_expand_down { pointerwin } {
  global view_dragged
  global view_x1
  global view_y1 
  global view_x2
  global view_y2 
  global workspace_height
  global cur_workspace
  global src_workspace
  global selection_owner
  
  set win $selection_owner
  if { $win == "" } return 
  
  set cur $cur_workspace($win)
  if { $view_dragged <= 0  &&  $src_workspace == $cur } {
    set x1 $view_x1
    set y1 $view_y1
    set height $workspace_height($cur)
    if { $y1 < $height  &&  $y1 >= 0 } {
      set x2 $view_x2
      set y2 [expr $view_y2 + 1]
      if { $y2 <= $height } {
        view_select $win $x1 $y1 $x2 $y2
        view_show_selection $win
        view_enter $pointerwin [winfo pointerx $win] [winfo pointery $win]
        status ""
      } else {
        status "Bottom edge reached"
      }
    }
  } else {
    if { $src_workspace != $cur } { reset_selection $win }
  }
}


#
# Initiate an action as a response to user pressing mouse button
# above some chip(s).
#
proc view_action { win x y { mode "auto" } } {
  global write_protect
  global mouse_mode
  global cur_workspace
  global workspace_chips
  global view_dragged

  # do we have a workspace
  set cur $cur_workspace($win)
  if { $cur < 0 } {
    # no valid workspace
    return 
  }

  # select operation according to mouse mode unless mode is supplied
  if { $mode == "auto" } {
    set mode $mouse_mode
  }

  # only select/pan/edit is allowed in write protected work space
  if { $write_protect($win)  \
      &&  $mode != "select"  &&  $mode != "pan"  &&  $mode != "edit" } {
    set mode "select"
  }

  # initiate action
  if { $mode == "select" } {    
    # start selecting chips
    view_select_xy $win $x $y
    set pos [view_at $win $x $y]
    if { $pos != "" } {
      set x [lindex $pos 0]
      set y [lindex $pos 1]
      set id $workspace_chips($cur,$x,$y)
      status "Chip #$id at $x,$y"
    }
  } elseif { $mode == "draw" } {
    # draw selection to under cursor
    undo_limit
    view_draw $win $x $y
  } elseif { $mode == "move" } {
    view_start_move $win $x $y    
  } elseif { $mode == "edit" } {
    # find out chip id
    view_select_xy $win $x $y
    set pos [view_at $win $x $y]
    if { $pos != "" } {
      set x [lindex $pos 0]
      set y [lindex $pos 1]
      set id $workspace_chips($cur,$x,$y)

      # edit chip in dialog
      if { $id > 0 } {
        edit_chip $id
      } else {
        status "Invalid chip"
      }
    }
    
    # prevent dragging and multi select
    set view_dragged -1
  } elseif { $mode == "pan" } {
    # pan view
    view_init_pan $win $x $y
  } else {
    modal_dialog "Error" "Invalid mode: $mode" error
  }
}


#
# Handle drag operation depnding on the mode.  In draw mode, paint chips
# to workspaces, in pan mode move the view etc.  Also, update mouse cursor,
# hint text and other visual clues to indicate that a drag operation is
# under way.
#
# Invoked by the user holding down some mouse button while moving mouse.
#
proc view_drag { win x y } {
  global view_dragged
  global chipw
  global chiph
  global view_drag_op
  
  if { $view_drag_op == "select" } {
    # initiate select
    if { $view_dragged == 0 } {
      status "Move mouse to select range of chips"
      $win configure -cursor left_ptr
      set view_dragged 1
    }

    # expand selection
    if { $view_dragged > 0 } {
      view_mouse_expand_selection $win $x $y
    }
    
  } elseif { $view_drag_op == "draw" } {
    # draw selection to under cursor
    view_draw $win $x $y
    
  } elseif { $view_drag_op == "pan" } {
    # pan view
    view_pan $win $x $y
    
  } elseif { $view_drag_op == "edit" } {
    # NOP
    
  } elseif { $view_drag_op == "move" } {
    if { $view_dragged > 0 } {
      # move view if necessary
      view_autoscroll $win $x $y
      view_highlight $win $x $y
    } elseif { $view_dragged == 0 } {
      # initiate drag
      set view_dragged 1
      view_highlight $win $x $y

      # tell user what to do
      if { $view_drag_op == "move" } {
        status "Release mouse button to move chips"
      }
    }
  }
}


#
# Release a grab, and perform operation on selected chips.  Invoked by the
# user releasing mouse button.
#
proc view_drop { win x y } {
  global view_fx
  global view_fy
  global view_x1
  global view_y1
  global view_x2
  global view_y2
  global view_dragged
  global cur_workspace
  global workspace_width
  global workspace_height
  global view_drag_op

  # get number of currently visible workspace in this widget
  set cur $cur_workspace($win)
  if { $cur < 0 } {
    # no valid workspace
    return 
  }

  # invoke operation depending on the drag mode
  if { $view_dragged > 0 } {
    if { $view_drag_op == "select" } {
      status "Chips selected"

    } elseif { $view_drag_op == "pan" } {
      # panning
      view_end_pan $win $x $y

    } elseif { $view_drag_op == "draw" } {
      status ""
      
    } elseif { $view_drag_op == "edit" } {
      status ""
      
    } elseif { $view_drag_op == "move" } {
      set pos [view_at $win $x $y]
      if { "$pos" != "" } {
        # cursor still over view
        set x1 $view_x1
        set y1 $view_y1
        set x2 $view_x2
        set y2 $view_y2
        set a1 [expr [lindex $pos 0] - $view_fx]
        set b1 [expr [lindex $pos 1] - $view_fy]
        set a2 [expr $a1 + ($x2 - $x1)]
        set b2 [expr $b1 + ($y2 - $y1)]

        # confine target to boundaries of the view
        if { 0 <= $a1  &&  $a2 <= $workspace_width($cur) \
            &&  0 <= $b1  &&  $b2 <= $workspace_height($cur) } {
          # Swap regions.  If the regions overlap the order of exchange
          # is important to preserve the relative location of the chips.
          if { $a1 != $x1  ||  $b1 != $y1 } {
            
            # cursor moved
            undo_limit
            if { $b1 < $y1  ||  ($b1 == $y1  &&  $a1 < $x1) } {
              for {set j $b1; set y $y1} {$j < $b2} {incr j; incr y} {
                for {set i $a1; set x $x1} {$i < $a2} {incr i; incr x} {
                  view_swap_chips $win $i $j $x $y
                }
              }
            } else {
              set j [expr $b2 - 1]
              set y [expr $y2 - 1]
              for {} {$j >= $b1} {incr j -1; incr y -1} {
                set i [expr $a2 - 1]
                set x [expr $x2 - 1]
                for {} {$i >= $a1} {incr i -1; incr x -1} {
                  view_swap_chips $win $i $j $x $y
                }
              }
            }
            
            # file has now been edited
            set_edited 1
            status "Chips relocated"
            
          } else {
            # user did not move the mouse
            status "Drag selection to move it"
          }

          # re-select the chips at the new location
          view_select $win $a1 $b1 $a2 $b2
        } else {
          # part of selection extends beyound the workspace
          bell
          status "Part of region extends outside the boundaries"
        }
      } else {
        # cursor not over a chip
        status "Cancel"
      }
    } else {
      modal_dialog "Error" "Bad drag operation: $view_drag_op" error
    }

    # remove highlight
    view_leave $win $x $y
  }

  # highlight chip under cursor
  view_enter $win $x $y

  # restore cursor
  restore_cursor $win

  # refuse to initiate drag after releasing the mouse button
  set view_drag_op "none"
  set view_dragged -1
}


#
# Set mouse cursor to match the current mouse mode.  The function is used
# after drag operations to give the user a visual clue on what the left
# mouse button does in the current mode.
#
proc restore_cursor { win } {
  global mouse_mode
  global write_protect

  if { $write_protect($win)  && \
      $mouse_mode!="pan" && $mouse_mode!="select" && $mouse_mode!="edit"} {
    # write protection prevents the mode from working
    $win configure -cursor left_ptr
  } else {
    # can edit workspace
    if { $mouse_mode == "select" } {
      $win configure -cursor left_ptr
    } elseif { $mouse_mode == "draw" } {
      $win configure -cursor pencil
    } elseif { $mouse_mode == "move" } {
      $win configure -cursor exchange
    } elseif { $mouse_mode == "pan" } {
      $win configure -cursor fleur
    } elseif { $mouse_mode == "edit" } {
      $win configure -cursor question_arrow
    }
  } 

  # compute mouse position in canvas
  set mx [winfo pointerx $win]
  set my [winfo pointery $win]

  # update highlight as the size of it depends on the mode
  if { $mx >= 0  &&  $my > 0 } {
    view_enter $win $mx $my
  }
}


#
# Initiate panning.
#
proc view_init_pan { win x y } {
  global view_dragged
  global view_drag_op
  global view_pan_x0
  global view_pan_y0
  global view_pan_xview
  global view_pan_yview

  # initialize for panning
  $win configure -cursor fleur
  set view_pan_x0 $x
  set view_pan_y0 $y

  # get coordinate of upper left coner of visible region
  set view_pan_xview [lindex [$win xview] 0]
  set view_pan_yview [lindex [$win yview] 0]

  # remove highlight
  $win delete "highlight"

  # set hint, and prevent them from changing automatically
  status "Move mouse to pan view"
  set view_dragged 1
  set view_drag_op "pan"
}


#
# Pan view.
#
proc view_pan { win x y } {
  global view_dragged
  global view_drag_op
  global view_pan_x0
  global view_pan_y0
  global view_pan_xview
  global view_pan_yview
  global chipw
  global chiph
  global workspace_width
  global workspace_height
  global cur_workspace
  global zoom
  global pan_speed
  global zoom_levels

  # get current zoom level
  set z [lindex $zoom_levels $zoom]

  # get id of currently visible workspace in this widget
  set cur $cur_workspace($win)
  
  if { $view_dragged  &&  $view_drag_op == "pan" } {
    # compute the distance moved
    set x [expr ($view_pan_x0 - $x) * $pan_speed]
    set y [expr ($view_pan_y0 - $y) * $pan_speed]

    # compute width and height of view in pixels
    set w [expr $workspace_width($cur)*$chipw*$z+1]
    set h [expr $workspace_height($cur)*$chiph*$z+1]

    # compute location of view
    set vx [expr $view_pan_xview + ($x+0.0)/$w]
    set vy [expr $view_pan_yview + ($y+0.0)/$h]
    if { $vx < 0 } { set vx 0 }
    if { $vy < 0 } { set vy 0 }
    
    # move view
    $win xview moveto $vx
    $win yview moveto $vy
  } else {
    modal_dialog "Error" "Not panning" error
  }
}


#
# End panning.
#
proc view_end_pan { win x y } {
  global view_dragged
  global view_drag_op

  # done with the panning
  set view_dragged -1
  set view_drag_op "none"
  status "Done panning"
}


#
# Select chip at coordinate (x,y) for further operations.  The function is
# typically invoked by user left-clicking a chip.
#
proc view_select_xy { win x y } {
  global view_dragged
  global view_x0
  global view_y0
  global view_drag_op

  # remove highlight
  view_leave $win $x $y

  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {
    # select chip under cursor
    view_select $win \
        [lindex $pos 0] [lindex $pos 1] \
        [expr [lindex $pos 0] + 1] [expr [lindex $pos 1] + 1]

    # Remember initial chip for drag operations.  This important if user
    # holds the mouse button down to select multiple chips.
    set view_x0 [lindex $pos 0]
    set view_y0 [lindex $pos 1]

    # drag mode
    set view_drag_op "select"

    # may drag the chip
    set view_dragged 0
  } else {
    # no chip under cursor, lose selection
    view_lose_selection $win
    
    # cannot initiate drag
    set view_dragged -1
  }
}


#
# Start dragging chips to different location.
#
proc view_start_move { win x y } {
  global view_dragged
  global view_fx
  global view_fy
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global view_drag_op
  global cur_workspace
  global src_workspace

  # get id of currently visible workspace in this widget
  set cur $cur_workspace($win)

  # remove highlight
  view_leave $win $x $y

  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {
    # chip under cursor
    set x0 [lindex $pos 0]
    set y0 [lindex $pos 1]
    set x1 $view_x1
    set y1 $view_y1
    set x2 $view_x2
    set y2 $view_y2    
    if { $x1 <= $x0  &&  $x0 < $x2  &&  $y1 <= $y0  &&  $y0 < $y2 \
         && $cur == $src_workspace } {
      # start dragging current selection
      set view_fx [expr $x0 - $x1]
      set view_fy [expr $y0 - $y1]
    } else {
      # cursor outside selection, select current chip
      set view_fx 0
      set view_fy 0
      view_select $win \
          [lindex $pos 0] [lindex $pos 1] \
          [expr [lindex $pos 0] + 1] [expr [lindex $pos 1] + 1]
    }

    # what to do with the chips
    set view_drag_op "move"

    # allow dragging
    set view_dragged 0
    view_drag $win $x $y
  } else {
    # no chip under cursor, lose selection
    view_lose_selection $win

    # disable drag
    set view_dragged -1
  }
}

#
# Paste selection to (x,y)
#
proc view_draw { win x y } {
  global view_x1
  global view_y1
  global view_x2
  global view_y2
  global view_dragged
  global cur_workspace
  global src_workspace
  global workspace_width
  global workspace_height
  global view_drag_op

  # get id of currently visible workspace in this widget
  set cur $cur_workspace($win)

  # Here, we could move the view to bring more chips into screen by
  # uncommenting the line below.  However, I personally do not like moving
  # the view when I am drawing maps, especially since there is no undo
  # at the moment.  Hence, I decided to comment the feature.
  ##view_autoscroll $win $x $y

  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {   
    # cursor still over view
    set x1 $view_x1
    set y1 $view_y1
    set x2 $view_x2
    set y2 $view_y2
    set a1 [lindex $pos 0]
    set b1 [lindex $pos 1]
    set a2 [expr $a1 + ($x2 - $x1)]
    set b2 [expr $b1 + ($y2 - $y1)]

    # confine target to boundaries of the workspace
    if { $a1 < 0 } {
      set a1 0
    }
    if { $b1 < 0 } {
      set b1 0
    }
    if { $a2 > $workspace_width($cur) } {
      set a2 $workspace_width($cur)
    }
    if { $b2 > $workspace_height($cur) } {
      set b2 $workspace_height($cur)
    }

    # Draw chips.  If the regions overlap, the processing order is important
    # for preserving the relative location of the chips.
    if { $a1 != $x1  ||  $b1 != $y1  ||  $cur != $src_workspace } {
      # mouse pointer is not directly over selection
      if { $b1<$y1 || ($b1==$y1 && $a1<$x1) || $cur!=$src_workspace} {
        # copy from left to right, top to bottom
        for {set j $b1; set y $y1} {$j < $b2} {incr j; incr y} {
          for {set i $a1; set x $x1} {$i < $a2} {incr i; incr x} {
            view_draw_chip $win $i $j $x $y
          }
        }
      } else {
        # copy from bottom to top, from right to left
        set j [expr $b2 - 1]
        set y [expr $y2 - 1]
        for {} {$j >= $b1} {incr j -1; incr y -1} {
          set i [expr $a2 - 1]
          set x [expr $x2 - 1]
          for {} {$i >= $a1} {incr i -1; incr x -1} {
            view_draw_chip $win $i $j $x $y
          }
        }
      }

      # file has now been edited
      set_edited 1
    }

    # Change cursor to pen to indicate that chips are being drawn.  For
    # some reason changing cursor shape is extremely heavy operation on
    # X11, so the cursor should never be changed unless absolutely
    # needed.  Perhaps the behaviour has something to do with the motion
    # tracking.
    if { [$win cget -cursor] != "pencil" } {
      $win configure -cursor pencil
    }

    # set up dragging
    set view_dragged 1
    set view_drag_op "draw"
  } else {
    # cursor not over a chip
    if { [$win cget -cursor] != "pirate" } {
      $win configure -cursor pirate
    }
  }

  # remove highlight until user releases mouse button
  view_leave $win $x $y
}


#
# Expand chip selection by dragging.  This is the same as
# view_expand_selection expect the canvas is scrolled automatically.
#
proc view_mouse_expand_selection { win x y } {
  view_autoscroll $win $x $y
  view_expand_selection $win $x $y
}


#
# Expand current selection to chip at (x,y).  Note that the function
# is only called as a result of a drag, so it must not touch the dragged
# state.  Especially, the function must not modify view_dragged
# variable.
#
proc view_expand_selection { win x y } {
  global view_x0
  global view_y0
  global view_x1
  global view_y1 
  global view_x2
  global view_y2 
  
  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {
    # a chip found
    if { $view_x1 >= 0 } {
      # some chips selected; enlarge selection
      set x1 $view_x0
      set y1 $view_y0
      set x2 [lindex $pos 0]
      set y2 [lindex $pos 1]

      # reorder coordinate such that (x1,y1) is always at upper left corner
      if { $x1 > $x2 } {
        set t  $x1
        set x1 $x2
        set x2 $t
      }
      if {$y1 > $y2} {
        set t  $y1
        set y1 $y2
        set y2 $t
      }
      incr x2
      incr y2

      # select chips in the enlarged region
      view_select $win $x1 $y1 $x2 $y2
    } else {
      # no other chips selected; select chip under cursor
      view_select $win \
          [lindex $pos 0] [lindex $pos 1] \
          [expr [lindex $pos 0] + 1] [expr [lindex $pos 1] + 1]
    }
  }
}


#
# Select chips (x1,y1)-(x2,y2) for further operations.  Note that the
# function does not ensure that the chips selected are in view.  If 
# necessary, call view_show_selection for that.
#
proc view_select { win x1 y1 x2 y2 } {
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global cur_workspace
  global src_workspace
  global selection_owner
  global view_border

 # get border width
  set bw $view_border($win)

  # remove highlight from previously selected chip
  if { $selection_owner != "" } {
    $selection_owner delete "select"
  }
 
  # remember selection
  set view_x1 $x1
  set view_y1 $y1
  set view_x2 $x2
  set view_y2 $y2
  set src_workspace $cur_workspace($win)
  set selection_owner $win

  # get width & height of the selection
  set width  [expr $x2 - $x1]
  set height [expr $y2 - $y1]

  # get location of the chip upper left chip
  set pos [view_pos $win $x1 $y1 $width $height]
  set x1p [expr [lindex $pos 0] - 2 + $bw]
  set y1p [expr [lindex $pos 1] - 2 + $bw]
  set x2p [expr [lindex $pos 2] + 2]
  set y2p [expr [lindex $pos 3] + 2]

  # highlight selected chip
  $win create rectangle $x1p $y1p $x2p $y2p \
      -tag "select" -width 4 -outline red
}


#
# Select something so that the user can move cursor.  The function is called
# when the selection is in some other workspace, and user wants to move
# the cursor.
#
proc reset_selection { win } {
  view_select $win 0 0 1 1
}


#
# Erase current chip selection.
#
proc view_lose_selection { win } {
  global view_x0
  global view_y0
  global view_fx
  global view_fy
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global src_workspace
  global cur_workspace

  # remove highlight from previously selected chip
  $win delete "select"
  set src_workspace $cur_workspace($win)

  # reset selection
  set view_x1 -1
  set view_x2 -1
  set view_y1 -1
  set view_y2 -1
  set view_x0 -1
  set view_x0 -1
  set view_fx -1
  set view_fx -1
}


#
# Mouse cursor enters a chip.
#
proc view_enter { win x y } {
  global view_x0
  global view_y0
  global view_x1
  global view_y1 
  global view_x2
  global view_y2
  global mouse_mode
  global write_protect
  global view_selection
  global view_border

  # if no chips are either operated or selected, then select nothing
  if { $mouse_mode == "pan" } {
    view_leave $win $x $y
    return
  }
  
  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {    
    # find index to the selected chip
    set a [lindex $pos 0]
    set b [lindex $pos 1]

    # get width and height of highlight
    if { $mouse_mode == "draw"  &&  ! $write_protect($win) } {
      # operation can affect several chips
      set w [expr $view_x2 - $view_x1]
      set h [expr $view_y2 - $view_y1]
      if { $w < 1} { set w 0 }
      if { $h < 1} { set h 0 }
    } else {
      # left click operates only a single chips
      set w 1
      set h 1
    }

    # get chip coordinate
    set pos [view_pos $win $a $b $w $h]

    # only draw selection if the coordinates are changes (improves speed)
    if { $pos != $view_selection } {
      # Highlight selected chip.  Note that we must be very careful here not
      # to trigger leave event for the chip: doing that would erase the
      # highlight, and consequentally again trigger this very event 
      # resulting to an endless loop.  That is why we must bind the leave
      # event to "highlight" class and *not* the "chip" class.
      set bw $view_border($win)
      set x1 [lindex $pos 0]
      set y1 [lindex $pos 1]
      set x2 [lindex $pos 2]
      set y2 [lindex $pos 3]
      $win delete "highlight"
      $win create rectangle [expr $x1+$bw] [expr $y1+$bw] $x2 $y2 \
          -tag "highlight" -stipple gray50 -fill yellow -width 0
      $win raise "highlight" "chip"
      set view_selection $pos
    }
  } else {
    # cursor not above any chip
    view_leave $win $x $y
  }
}


#
# Mouse cursor leaves a chip.
#
proc view_leave { win x y } {
  global view_selection
  
  # remove highlight from any item
  $win delete "highlight"
  set view_selection ""
}


#
# Auto-scroll window to bring more chips into view.  Used while dragging.
#
proc view_autoscroll { win x y } {
  global chipw
  global chiph
  global autoscroll_margin

  # Define scroll region.  Moving mouse pointer outside this area triggers
  # scrolling.
  set x1 [expr [winfo rootx $win] + $autoscroll_margin]
  set y1 [expr [winfo rooty $win] + $autoscroll_margin]
  set x2 [expr [winfo rootx $win] + [winfo width $win] -$autoscroll_margin]
  set y2 [expr [winfo rooty $win] + [winfo height $win]-$autoscroll_margin]

  # scroll to bring more chips into view
  if { $x < $x1 } {
    if { [lindex [$win xview] 0] > 0 } {
      $win xview scroll -1 units
    } else {
      $win xview moveto 0
    }
  } elseif { $y < $y1 } {
    if { [lindex [$win yview] 0] > 0 } {
      $win yview scroll -1 units
    } else {
      $win yview moveto 0
    }
  }
  if { $x > $x2 } {
    $win xview scroll 1 units
  } elseif { $y > $y2 } {
    $win yview scroll 1 units
  }
}


#
# Ensure that selected chips are in view.  This is mostly needed when
# selecting chips with keyboard.
#
proc view_show_selection { win } {
  global workspace_width
  global workspace_height
  global cur_workspace
  global view_x1
  global view_y1 
  global view_x2
  global view_y2 

  # get id of current workspace
  set cur $cur_workspace($win)

  # get dimensions of current workspace
  set wt $workspace_width($cur)
  set ht $workspace_height($cur)
  
  # compute location of selection within the view
  set selx1 [expr ($view_x1+0.0) / $wt]
  set sely1 [expr ($view_y1+0.0) / $ht]
  set selx2 [expr ($view_x2+0.0) / $wt]
  set sely2 [expr ($view_y2+0.0) / $ht]

  # find out visible part of canvas
  set viewx1 [lindex [$win xview] 0]
  set viewy1 [lindex [$win yview] 0]
  set viewx2 [lindex [$win xview] 1]
  set viewy2 [lindex [$win yview] 1]

  # adjust visible region
  if { $selx2 > $viewx2 } {
    # show right edge of selection
    set viewx1 [expr $selx2 - ($viewx2 - $viewx1)]
  } elseif { $selx1 < $viewx1 } {
    if { [expr $selx2 - $selx1] < [expr $viewx2 - $viewx1] } {
      # just a few chips selected
      set viewx1 $selx1
    } else {
      # selection is wider than window
      set viewx1 [expr $selx2 - ($viewx2 - $viewx1)]      
    }
  }
  if { $sely2 > $viewy2 } {
    set viewy1 [expr $sely2 - ($viewy2 - $viewy1)]
  } elseif { $sely1 < $viewy1 } {
    if { [expr $sely2 - $sely1] < [expr $viewy2 - $viewy1] } {
      # just a few chips selected
      set viewy1 $sely1
    } else {
      # selection is wider than window
      set viewy1 [expr $sely2 - ($viewy2 - $viewy1)]
    }
  }

  # refuse to reveal area outside the workspace
  if { $viewx1 < 0 } { set viewx1 0 }
  if { $viewy1 < 0 } { set viewy1 0 }

  # move canvas
  $win xview moveto $viewx1
  $win yview moveto $viewy1
}


#
# Highlight drop area affected by a drag operation.  The function is only
# used with move operation.
#
proc view_highlight { win x y } {
  global view_fx
  global view_fy
  global view_x1
  global view_y1 
  global view_x2
  global view_y2 
  global cur_workspace
  global workspace_width
  global workspace_height

  # get number of visible workspace
  set cur $cur_workspace($win)

  # find out chip under cursor
  set pos [view_at $win $x $y]
  if { "$pos" != "" } {
    # erase existing highlight
    view_leave $win $x $y
    
    # find index to the selected chip
    set a [lindex $pos 0]
    set b [lindex $pos 1]

    # get width & height of the selection
    set a [expr $a - $view_fx]
    set b [expr $b - $view_fy]
    set width  [expr $view_x2 - $view_x1]
    set height [expr $view_y2 - $view_y1]

    # is within range?
    if { 0<=$a && [expr $a+$width] <= $workspace_width($cur) \
        &&  0<=$b && [expr $b+$height] <= $workspace_height($cur) } {
      # get location of the chip
      set pos [view_pos $win $a $b $width $height]
      set x1 [lindex $pos 0]
      set y1 [lindex $pos 1]
      set x2 [lindex $pos 2]
      set y2 [lindex $pos 3]

      # highlight selected chips
      $win create rectangle $x1 $y1 [expr $x2+1] [expr $y2+1] \
          -tag "highlight" -stipple gray50 -fill blue \
          -width 0 -outline black
      $win raise "highlight" "chip"

      # change cursor to indicate successful drop
      $win configure -cursor exchange
    } else {
      # cannot drop here (part of selection extends beyond boudaries)
      $win configure -cursor pirate
    }
  } else {
    # no chip under cursor
    view_leave $win $x $y
    $win configure -cursor pirate
  }
}


#
# Swap chips at (i,j) and (x,y).
# 
proc view_swap_chips { win i j x y } {
  global cur_workspace
  global workspace_chips
  global chip_image
  global zoom

  set cur $cur_workspace($win)
  set t $workspace_chips($cur,$i,$j)
  set_chip $cur $i $j $workspace_chips($cur,$x,$y)
  set_chip $cur $x $y $t
  
  # refresh chips on canvas
  set chip $workspace_chips($cur,$i,$j)
  $win itemconfigure "chip_${i}_${j}" -image $chip_image($chip,1)
  set chip $workspace_chips($cur,$x,$y)
  $win itemconfigure "chip_${x}_${y}" -image $chip_image($chip,1)
}


#
# Draw chip from (x,y) to (i,j).
# 
proc view_draw_chip { win i j x y } {
  global cur_workspace
  global workspace_chips
  global chip_image
  global zoom
  global src_workspace

  # get id of currently visible workspace in this widget
  set cur $cur_workspace($win)

  set chip $workspace_chips($src_workspace,$x,$y)
  if { $workspace_chips($cur,$i,$j) != $chip  &&  $chip != 0 } {
    # must update chip
    set_chip $cur $i $j $chip

    # refresh chips on canvas
    $win itemconfigure chip_${i}_${j} -image $chip_image($chip,1)
  }
}


#
# Get location of chip at screen coordinate (x,y).  The function is used
# for finding out chip under mouse cursor.  Hence, coordinates outside
# the view widget are not considered.
#
proc view_at { win x y } {
  global cur_workspace
  global zoom
  global chipw
  global chiph
  global workspace_width
  global workspace_height
  global view_border
  global zoom_levels

  # get current workspace
  set cur $cur_workspace($win)
  if { $cur < 0 } {
    # not a valid workspace
    return ""
  }

  # get border width
  set bw $view_border($win)
  
  # get current zoom level
  set z [lindex $zoom_levels $zoom]

  # get screen coordinates of the widget
  set x1 [winfo rootx $win]
  set y1 [winfo rooty $win]
  set x2 [expr $x1 + [winfo width $win]]
  set y2 [expr $y1 + [winfo height $win]]

  # is the location within the visible region?
  if { $x1 <= $x  &&  $x <= $x2  &&  $y1 <= $y  &&  $y <= $y2 } {
    # yes, compute cursor position within the scrollable canvas
    set x [expr int([$win canvasx $x] - [winfo rootx $win])]
    set y [expr int([$win canvasy $y] - [winfo rooty $win])]

    # compute index to workspace_chips array
    set a [expr int(($x - 2) / ($chipw * $z + $bw))]
    set b [expr int(($y - 2) / ($chiph * $z + $bw))]

    # see if coordinate (x,y) is over a chip
    if { 0 <= $a  &&  $a < $workspace_width($cur) \
        &&  0 <= $b  &&  $b < $workspace_height($cur) } {
      # return chip index
      return "$a $b"
    } else {
      # coordinate (x,y) not touching a chip (small workspace)
      return ""
    }
  } else {
    # location (x,y) not over the widget
    return ""
  }
}


#
# Compute window coordinates of a chips at (a,b) - (a+w,b+h).
#
proc view_pos { win a b {w 1} {h 1} } {
  global zoom
  global chipw
  global chiph
  global view_border
  global zoom_levels

  # get border width
  set bw $view_border($win)

  # get current zoom level
  set z [lindex $zoom_levels $zoom]

  # chip width in pixels with border
  set cw [expr $chipw * $z + $bw]
  set ch [expr $chiph * $z + $bw]

  # position of chip in pixel coordinates
  set x1 [expr $a * $cw]
  set y1 [expr $b * $ch]
  set x2 [expr $x1 + $w * $cw]
  set y2 [expr $y1 + $h * $ch]
  return "$x1 $y1 $x2 $y2"
}
